from .info import (
    __appname__,
    __author__,
    __contact__,
    __description__,
    __descriptionfull__,
    __license__,
    __licensefull__,
    __maintainer__,
    __maintainer_contact__,
    __packagename__,
    __projecturl__,
)
from .version import __version__
